import networkx as nx
import numpy as np
import pandas as pd
from nose import with_setup

from pybbn.learn.cb import get_mwst_skeleton, get_v_structures, MwstAlgo
from pybbn.learn.data import DiscreteData


def setup():
    """
    Setup.
    :return: None.
    """
    np.random.seed(37)


def teardown():
    """
    Teardown.
    :return: None.
    """


def get_good_df():
    """
    Gets a good dataframe.
    :return: Dataframe.
    """
    return pd.DataFrame({
        'x1': ['t', 't', 't', 't', 't', 'f', 'f', 'f', 'f', 'f'],
        'x2': ['t', 't', 't', 't', 'f', 't', 'f', 'f', 'f', 'f'],
        'x3': ['t', 't', 't', 't', 't', 't', 'f', 'f', 'f', 'f'],
        'x4': ['t', 't', 't', 't', 't', 'f', 'f', 'f', 'f', 'f']
    })


@with_setup(setup, teardown)
def test_get_mwst_skeleton():
    """
    Tests getting the MWST skeleton structure.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    g = get_mwst_skeleton(data)
    assert len(g.nodes) == 4


@with_setup(setup, teardown)
def test_get_v_structures():
    """
    Tests getting all the v-structures from an undirected graph.
    :return: None.
    """
    g = nx.Graph()
    g.add_node(0)
    g.add_node(1)
    g.add_node(2)
    g.add_node(4)
    g.add_edge(0, 1)
    g.add_edge(0, 2)
    g.add_edge(2, 3)

    v_structures = get_v_structures(g)
    assert len(v_structures) == 2
    assert (1, 0, 2) in v_structures
    assert (0, 2, 3) in v_structures


@with_setup(setup, teardown)
def test_fit():
    """
    Tests fitting data.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    mwst = MwstAlgo()
    mwst.fit(data)

    bbn = mwst.bbn
    assert 4 == len(bbn.nodes)
    assert 3 == len(bbn.edges)



import itertools

import networkx as nx
import numpy as np
from networkx.algorithms.dag import is_directed_acyclic_graph

from pybbn.graph.dag import Bbn
from pybbn.graph.edge import EdgeType, Edge
from pybbn.graph.node import BbnNode
from pybbn.graph.variable import Variable


class MwstAlgo(object):
    """
    Maximum weight spanning tree algorithm. Also known as Chow-Liu algorithm.
    """
    def __init__(self):
        """
        Ctor.
        """
        self.data = None
        self.bbn = None

    def fit(self, data):
        """
        Learns the structure and parameter of a Bayesian belief network from the data.
        :param data: Data.
        :return: BBN.
        """
        g = self.learn_structure(data)
        p = self.learn_parameters(data, g)

        bbn = build_bbn(data.variable_profiles, g, p)

        self.data = data
        self.bbn = bbn

    def learn_parameters(self, data, g):
        """
        Learns the parameters (conditional probability tables, local probability models).
        :param data: Data.
        :param g: DAG.
        :return: A dictionary associated with nodes/variables and their parameters.
        """
        cpts = {}
        variable_profiles = data.variable_profiles
        nodes = list(g.nodes)
        for idx in nodes:
            child = g.nodes[idx]['name']
            c_values = variable_profiles[child]

            parents = list(g.predecessors(idx))
            parents = [g.nodes[pa_id]['name'] for pa_id in parents]
            p_values = list(itertools.product(*[variable_profiles[name] for name in parents]))

            if len(parents) == 0:
                cpt = [data.__get_prob__(child, value) for value in c_values]
            else:
                cpt = []
                for values in p_values:
                    for value in c_values:
                        p = data.__get_cond_prob_set__(child, value, parents, values)
                        cpt.append(p)
            cpts[idx] = cpt

        return cpts

    def learn_structure(self, data):
        """
        Learns the BBN structure.
        :param data: Data.
        :return: DAG.
        """
        # gets the skeleton, which is a tree, undirected graph
        u = get_mwst_skeleton(data)

        # gets all the v-structures, if any
        variables = data.get_variables(by_name=False)
        v_structures = []
        for v_structure in get_v_structures(u):
            name1 = variables[v_structure[0]]
            name2 = variables[v_structure[2]]
            names = [variables[v_structure[1]]]
            cmi, result = data.is_cond_dep(name1, name2, names)
            if result is True:
                t = (v_structure[0], v_structure[1], v_structure[2], cmi)
                v_structures.append(t)
        v_structures = sorted(v_structures, key=lambda t: t[3], reverse=True)

        # create the DAG, this graph will have edges oriented
        g = nx.DiGraph()
        for idx in u.nodes:
            g.add_node(idx, name=u.nodes[idx]['name'])

        # attempt to orient edges by v-structures
        for v_structure in v_structures:
            left_node = v_structure[0]
            mid_node = v_structure[1]
            right_node = v_structure[2]

            edge1 = (left_node, mid_node, {})
            edge2 = (right_node, mid_node, {})

            g.add_edges_from([edge1, edge2])

            if is_directed_acyclic_graph(g) is False:
                g.remove_edges_from([edge1, edge2])

        # attempt to orient by likelihood
        # FIXME: note this can create v-structures
        missing_edges = get_missing_edges(u, g)
        directed_edges = get_likely_directed_edges(missing_edges, data)
        for edge in directed_edges:
            pa_node = edge[0]
            ch_node = edge[1]
            dedge = (pa_node, ch_node, {})
            g.add_edges_from([dedge])

            if is_directed_acyclic_graph(g) is False:
                g.remove_edges_from([dedge])

        # now just orient randomly, so long as the graph is a dag
        missing_edges = get_missing_edges(u, g)
        while len(missing_edges) != 0:
            for edge in missing_edges:
                pa_id = np.random.randint(0, 2)
                ch_id = 1 if pa_id == 0 else 0

                pa_node = edge[pa_id]
                ch_node = edge[ch_id]

                dedge = (pa_node, ch_node, {})
                g.add_edges_from([dedge])

                if is_directed_acyclic_graph(g) is False:
                    g.remove_edges_from([dedge])

            missing_edges = get_missing_edges(u, g)

        return g


def build_bbn(variable_profiles, g, p):
    """
    Builds a BBN from a DAG, g, and paremeters, p.
    :param variable_profiles: Variable profiles.
    :param g: DAG.
    :param p: Parameters.
    :return: BBN.
    """
    bbn = Bbn()
    nodes = list(g.nodes)
    bbn_node_dict = {}
    for idx in nodes:
        name = g.nodes[idx]['name']
        domain = variable_profiles[name]
        cpt = p[idx]

        v = Variable(idx, name, domain)
        n = BbnNode(v, cpt)
        bbn.add_node(n)
        bbn_node_dict[idx] = n

    edges = list(g.edges)
    for edge in edges:
        pa = bbn_node_dict[edge[0]]
        ch = bbn_node_dict[edge[1]]
        e = Edge(pa, ch, EdgeType.DIRECTED)
        bbn.add_edge(e)

    return bbn


def get_likely_directed_edges(edges, data):
    """
    Gets directed edged based on likelihood for each undirected edge.
    :param edges: List of edges.
    :param data: Data.
    :return: List of directed edge.
    """
    directed_edges = []
    variables = data.get_variables(by_name=False)
    for edge in edges:
        v1 = variables[edge[0]]
        v2 = variables[edge[1]]

        v1_v2 = data.get_local_kutato(v1, [v2])
        v2_v1 = data.get_local_kutato(v2, [v1])
        if v1_v2 > v2_v1:
            t = (edge[0], edge[1])
            directed_edges.append(t)
        elif v1_v2 < v2_v1:
            t = (edge[1], edge[0])
            directed_edges.append(t)
    return directed_edges


def get_missing_edges(u, g):
    """
    Gets the missing edges in g that is in u.
    :param u: Undirected graph.
    :param g: Directed acyclic graph.
    :return: List of tuples, where each tuple is a pair of nodes designating an edge.
    """
    edges = []
    for edge in list(u.edges):
        n1 = edge[0]
        n2 = edge[1]

        if n1 not in list(g.adj[n2]) and n2 not in list(g.adj[n1]):
            t = (n1, n2)
            edges.append(t)
    return edges


def get_v_structures(g):
    """
    Gets all the v-structures in g. For three nodes, a, b, c, where a is connected to b,
    b is connected c, and a and c are not connected, this configuration is called a v-structure.
    :param g: Undirected graph.
    :return: List of tuples representing v-structure.
    """
    v_structures = []

    nodes = list(g.nodes)
    for node in nodes:
        neighbors = list(g.adj[node])
        for ne1 in neighbors:
            for ne2 in neighbors:
                if ne1 < ne2 and ne1 not in g.adj[ne2] and ne2 not in g.adj[ne1]:
                    t = (ne1, node, ne2)
                    v_structures.append(t)

    return v_structures


def get_mwst_skeleton(data):
    """
    Gets the undirected graph that represents the maximum weight spanning tree (MWST)
    of the data.
    :param data: Data.
    :return: Undirected graph (MWST).
    """
    g = nx.Graph()
    variables = data.get_variables()
    for k, v in variables.items():
        g.add_node(v, name=k)

    mis = data.get_pairwise_mutual_information()
    for mi in mis:
        lhs_node = variables[mi[0]]
        rhs_node = variables[mi[1]]
        t = (lhs_node, rhs_node, {})
        g.add_edges_from([t])

        if len(g.edges) == len(variables) - 1:
            break

    return g



from pybbn.graph.edge import JtEdge
from pybbn.graph.jointree import JoinTree
from pybbn.graph.node import SepSet


class Transformer(object):
    """
    Transformer. Transforms a list of cliques into a join tree.
    """

    @staticmethod
    def transform(cliques):
        """
        Transforms the cliques into a join tree.
        :param cliques: List of cliques.
        :return: Join tree.
        """
        join_tree = JoinTree()
        sep_sets = Transformer.get_sep_sets(cliques)
        n = (len(cliques) - 1) * 2
        total = 0

        for i in range(len(sep_sets)):
            join_tree.add_edge(JtEdge(sep_sets[i]))
            total += 2
            if total == n:
                break

        return join_tree

    @staticmethod
    def get_sep_sets(cliques):
        """
        Gets all pair-wise separation-sets.
        :param cliques: Array of cliques.
        :return: Array of separation sets sorted descendingly by mass followed by
        cost (asc) and id (asc).
        """
        sep_sets = []
        size = len(cliques)
        for i in range(size):
            for j in range(i + 1, size):
                sep_set = SepSet(cliques[i], cliques[j])
                sep_sets.append(sep_set)
        return sorted(sep_sets, key=lambda x: (-1 * x.mass, x.cost, x.id))



import json

from nose import with_setup

from pybbn.graph.factory import Factory
from pybbn.pptc.inferencecontroller import InferenceController


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


def get_json_string():
    return json.dumps(get_dict())


def get_dict():
    return {
        "V": ["Letter", "Grade", "Intelligence", "SAT", "Difficulty"],
        "E": [["Difficulty", "Grade"],
              ["Intelligence", "Grade"],
              ["Intelligence", "SAT"],
              ["Grade", "Letter"]],
        "Vdata": {
            "Letter": {
                "ord": 4,
                "numoutcomes": 2,
                "vals": ["weak", "strong"],
                "parents": ["Grade"],
                "children": None,
                "cprob": {
                    "['A']": [.1, .9],
                    "['B']": [.4, .6],
                    "['C']": [.99, .01]
                }
            },
            "SAT": {
                "ord": 3,
                "numoutcomes": 2,
                "vals": ["lowscore", "highscore"],
                "parents": ["Intelligence"],
                "children": None,
                "cprob": {
                    "['low']": [.95, .05],
                    "['high']": [.2, .8]
                }
            },
            "Grade": {
                "ord": 2,
                "numoutcomes": 3,
                "vals": ["A", "B", "C"],
                "parents": ["Difficulty", "Intelligence"],
                "children": ["Letter"],
                "cprob": {
                    "['easy', 'low']": [.3, .4, .3],
                    "['easy', 'high']": [.9, .08, .02],
                    "['hard', 'low']": [.05, .25, .7],
                    "['hard', 'high']": [.5, .3, .2]
                }
            },
            "Intelligence": {
                "ord": 1,
                "numoutcomes": 2,
                "vals": ["low", "high"],
                "parents": None,
                "children": ["SAT", "Grade"],
                "cprob": [.7, .3]
            },
            "Difficulty": {
                "ord": 0,
                "numoutcomes": 2,
                "vals": ["easy", "hard"],
                "parents": None,
                "children": ["Grade"],
                "cprob": [.6, .4]
            }
        }
    }


@with_setup(setup, teardown)
def test_from_libpgm_discrete_json():
    """
    Tests create py-bbn BBN from JSON string specifying libpgm BBN.
    :return: None.
    """
    j = get_json_string()
    bbn = Factory.from_libpgm_discrete_json(j)

    assert len(bbn.nodes) == 5
    assert len(bbn.edges) == 4


@with_setup(setup, teardown)
def test_from_libpgm_discrete_dictionary():
    """
    Tests create py-bbn BBN from dictionary specifying libpgm BBN.
    :return: None.
    """
    d = get_dict()
    bbn = Factory.from_libpgm_discrete_dictionary(d)

    assert len(bbn.nodes) == 5
    assert len(bbn.edges) == 4

    # FIXME this is not working right somehow
    join_tree = InferenceController.apply(bbn)

    for node in join_tree.get_bbn_nodes():
        potential = join_tree.get_bbn_potential(node)
        print(node)
        print(potential)
        print('>')




from pybbn.pptc.evidencecollector import EvidenceCollector
from pybbn.pptc.evidencedistributor import EvidenceDistributor


class Propagator(object):
    """
    Evidence propagator.
    """

    @staticmethod
    def propagate(join_tree):
        """
        Propagates evidence.
        :param join_tree: Join tree.
        :return: Join tree.
        """
        cliques = sorted(join_tree.get_cliques(), key=lambda c: c.id)
        x = cliques[0]
        # print(x)

        join_tree.unmark_cliques()
        Propagator.collect_evidence(join_tree, x)

        join_tree.unmark_cliques()
        Propagator.distribute_evidence(join_tree, x)

        return join_tree

    @staticmethod
    def collect_evidence(join_tree, start):
        """
        Collects evidence.
        :param join_tree: Join tree.
        :param start: Start clique.
        """
        collector = EvidenceCollector(join_tree, start)
        collector.start()

    @staticmethod
    def distribute_evidence(join_tree, start):
        """
        Distributes evidence.
        :param join_tree: Join tree.
        :param start: Start clique.
        """
        distributor = EvidenceDistributor(join_tree, start)
        distributor.start()



from nose import with_setup

from pybbn.graph.dag import BbnUtil
from pybbn.pptc.potentialinitializer import PotentialInitializer


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_potential_initializer():
    """
    Tests potential initialization.
    :return: None.
    """
    bbn = BbnUtil.get_huang_graph()
    PotentialInitializer.init(bbn)

    # assert later
    # for node in bbn.get_nodes():
    #     potential = node.potential
    #     print('{} {}'.format(node.id, node.variable.name))
    #     print(potential)
    #
    # assert 1 == 2



import pandas as pd
from nose import with_setup
from nose.tools import assert_almost_equal

from pybbn.learn.data import DiscreteData


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


def get_good_df():
    """
    Gets a good dataframe.
    :return: Dataframe.
    """
    return pd.DataFrame({
        'x1': ['t', 't', 't', 't', 't', 'f', 'f', 'f', 'f', 'f'],
        'x2': ['t', 't', 't', 't', 'f', 't', 'f', 'f', 'f', 'f'],
        'x3': ['t', 't', 't', 't', 't', 't', 'f', 'f', 'f', 'f'],
        'x4': ['t', 't', 't', 't', 't', 'f', 'f', 'f', 'f', 'f']
    })


def get_bad_df():
    """
    Tests a bad dataframe.
    :return: Dataframe
    """
    return pd.DataFrame({
        'x1': ['t', 't', 't', 't', 't', 'f', 'f', 'f', 'f', 'f'],
        'x2': ['t', 't', 't', 't', 'f', 't', 'f', 'f', 'f', 'f'],
        'x3': ['t', 't', 't', 't', 't', 't', 'f', 'f', 'f', 'f'],
        'x4': ['t', 't', 't', 't', 't', 't', 't', 't', 't', 't']
    })


@with_setup(setup, teardown)
def test_count():
    """
    Tests simple counting.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    assert 5 == data.__count__(['x1'], ['t'])
    assert 5 == data.__count__(['x1'], ['f'])

    assert 5 == data.__count__(['x2'], ['t'])
    assert 5 == data.__count__(['x2'], ['f'])

    assert 6 == data.__count__(['x3'], ['t'])
    assert 4 == data.__count__(['x3'], ['f'])

    assert 5 == data.__count__(['x4'], ['t'])
    assert 5 == data.__count__(['x4'], ['f'])


@with_setup(setup, teardown)
def test_count_parents_child():
    """
    Tests counting parent-child values.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert 4 == data.__count_parents_child__('x1', 't', ['x2'], ['t'])
    assert 1 == data.__count_parents_child__('x1', 't', ['x2'], ['f'])
    assert 1 == data.__count_parents_child__('x1', 'f', ['x2'], ['t'])
    assert 4 == data.__count_parents_child__('x1', 'f', ['x2'], ['f'])


@with_setup(setup, teardown)
def test_get_local_kutato():
    """
    Tests computing local Kutato score.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(-1.540639, data.get_local_kutato('x1', ['x2']), places=5)
    assert_almost_equal(-1.540639, data.get_local_kutato('x2', ['x1']), places=5)

    assert_almost_equal(-1.082607, data.get_local_kutato('x1', ['x3']), places=5)
    assert_almost_equal(-1.124863, data.get_local_kutato('x3', ['x1']), places=5)

    assert_almost_equal(-0.709086, data.get_local_kutato('x1', ['x4']), places=5)
    assert_almost_equal(-0.709086, data.get_local_kutato('x4', ['x1']), places=5)


@with_setup(setup, teardown)
def test_is_valid_good():
    """
    Tests if a dataset is valid with a good dataframe.
    :return: None.
    """
    try:
        df = get_good_df()
        data = DiscreteData(df)
        assert 1 == 1
    except Exception:
        assert 1 == 2


@with_setup(setup, teardown)
def test_is_valid_bad():
    """
    Tests if a dataset is valid with a bad dataframe.
    :return: None.
    """
    try:
        df = get_bad_df()
        data = DiscreteData(df)
        assert 1 == 2
    except Exception:
        assert 1 == 1


@with_setup(setup, teardown)
def test_get_variable_profiles():
    """
    Tests getting variable profiles.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    profiles = data.get_variable_profiles()

    print(data)
    assert 4 == len(profiles)

    for i in range(4):
        name = 'x{}'.format(i + 1)
        assert name in profiles

        profile = profiles[name]
        assert 't' in profile
        assert 'f' in profile


@with_setup(setup, teardown)
def test_sort():
    """
    Tests sorting names and values by names ascending.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    observed = data.__sort__(['a', 'c', 'b', 'd'], ['t', 'f', 't', 'f'])
    expected = [('a', 't'), ('b', 't'), ('c', 'f'), ('d', 'f')]
    assert len(observed) == len(expected)
    for o, e in zip(observed, expected):
        assert o[0] == e[0]
        assert o[1] == e[1]


@with_setup(setup, teardown)
def test_to_query():
    """
    Tests generating query.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    names = ['a', 'c', 'b', 'd']
    vals = ['t', 'f', 't', 'f']
    observed = data.__to_query__(names, vals)
    expected = 'a == "t" and b == "t" and c == "f" and d == "f"'
    assert observed == expected


@with_setup(setup, teardown)
def test_get_prob():
    """
    Gets getting simple probability.
    :return: None.
    """
    data = DiscreteData(get_good_df())
    assert_almost_equal(0.5, data.__get_prob__('x1', 't'))
    assert_almost_equal(0.5, data.__get_prob__('x1', 'f'))
    assert_almost_equal(0.5, data.__get_prob__('x2', 't'))
    assert_almost_equal(0.5, data.__get_prob__('x2', 'f'))
    assert_almost_equal(0.6, data.__get_prob__('x3', 't'))
    assert_almost_equal(0.4, data.__get_prob__('x3', 'f'))
    assert_almost_equal(0.5, data.__get_prob__('x4', 't'))
    assert_almost_equal(0.5, data.__get_prob__('x4', 'f'))


@with_setup(setup, teardown)
def test_get_joint_prob():
    """
    Tests getting joint probability between two variables.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.4, data.__get_joint_prob__('x1', 't', 'x2', 't'))
    assert_almost_equal(0.1, data.__get_joint_prob__('x1', 't', 'x2', 'f'))
    assert_almost_equal(0.1, data.__get_joint_prob__('x1', 'f', 'x2', 't'))
    assert_almost_equal(0.4, data.__get_joint_prob__('x1', 'f', 'x2', 'f'))


@with_setup(setup, teardown)
def test_get_joint_prob_set():
    """
    Tests getting joint probability between a variable and a set of variables.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.4, data.__get_joint_prob_set__('x1', 't', ['x2', 'x3'], ['t', 't']))
    assert_almost_equal(0.0, data.__get_joint_prob_set__('x1', 't', ['x2', 'x3'], ['t', 'f']))
    assert_almost_equal(0.1, data.__get_joint_prob_set__('x1', 't', ['x2', 'x3'], ['f', 't']))
    assert_almost_equal(0.0, data.__get_joint_prob_set__('x1', 't', ['x2', 'x3'], ['f', 'f']))

    assert_almost_equal(0.1, data.__get_joint_prob_set__('x1', 'f', ['x2', 'x3'], ['t', 't']))
    assert_almost_equal(0.0, data.__get_joint_prob_set__('x1', 'f', ['x2', 'x3'], ['t', 'f']))
    assert_almost_equal(0.0, data.__get_joint_prob_set__('x1', 'f', ['x2', 'x3'], ['f', 't']))
    assert_almost_equal(0.4, data.__get_joint_prob_set__('x1', 'f', ['x2', 'x3'], ['f', 'f']))


@with_setup(setup, teardown)
def test_get_joint_prob_triplet():
    """
    Tests getting joint probability between two variables and a set of variables.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.4, data.__get_joint_prob_triplet__('x1', 't', 'x2', 't', ['x3'], ['t']))
    assert_almost_equal(0.0, data.__get_joint_prob_triplet__('x1', 't', 'x2', 't', ['x3'], ['f']))
    assert_almost_equal(0.1, data.__get_joint_prob_triplet__('x1', 't', 'x2', 'f', ['x3'], ['t']))
    assert_almost_equal(0.0, data.__get_joint_prob_triplet__('x1', 't', 'x2', 'f', ['x3'], ['f']))

    assert_almost_equal(0.1, data.__get_joint_prob_triplet__('x1', 'f', 'x2', 't', ['x3'], ['t']))
    assert_almost_equal(0.0, data.__get_joint_prob_triplet__('x1', 'f', 'x2', 't', ['x3'], ['f']))
    assert_almost_equal(0.0, data.__get_joint_prob_triplet__('x1', 'f', 'x2', 'f', ['x3'], ['t']))
    assert_almost_equal(0.4, data.__get_joint_prob_triplet__('x1', 'f', 'x2', 'f', ['x3'], ['f']))


@with_setup(setup, teardown)
def test_get_cond_prob():
    """
    Tests getting conditional probability between two variables.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.8, data.__get_cond_prob__('x1', 't', 'x2', 't'))
    assert_almost_equal(0.2, data.__get_cond_prob__('x1', 't', 'x2', 'f'))
    assert_almost_equal(0.2, data.__get_cond_prob__('x1', 'f', 'x2', 't'))
    assert_almost_equal(0.8, data.__get_cond_prob__('x1', 'f', 'x2', 'f'))


@with_setup(setup, teardown)
def test_get_cond_prob_set():
    """
    Tests getting conditional probability between a variable and a set of variables.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.8, data.__get_cond_prob_set__('x1', 't', ['x2', 'x3'], ['t', 't']))
    assert_almost_equal(0.2, data.__get_cond_prob_set__('x1', 'f', ['x2', 'x3'], ['t', 't']))

    assert_almost_equal(0.0, data.__get_cond_prob_set__('x1', 't', ['x2', 'x3'], ['f', 'f']))
    assert_almost_equal(1.0, data.__get_cond_prob_set__('x1', 'f', ['x2', 'x3'], ['f', 'f']))

    assert_almost_equal(0.0, data.__get_cond_prob_set__('x1', 't', ['x2', 'x3'], ['t', 'f']))
    assert_almost_equal(0.0, data.__get_cond_prob_set__('x1', 'f', ['x2', 'x3'], ['t', 'f']))

    assert_almost_equal(1.0, data.__get_cond_prob_set__('x1', 't', ['x2', 'x3'], ['f', 't']))
    assert_almost_equal(0.0, data.__get_cond_prob_set__('x1', 'f', ['x2', 'x3'], ['f', 't']))


@with_setup(setup, teardown)
def test_get_cond_prob_triplet():
    """
    Tests getting conditional probability between two variables and a third set of variables.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.6666666666666667, data.__get_cond_prob_triplet__('x1', 't', 'x2', 't', ['x3'], ['t']))
    assert_almost_equal(0.0, data.__get_cond_prob_triplet__('x1', 't', 'x2', 't', ['x3'], ['f']))
    assert_almost_equal(0.16666666666666669, data.__get_cond_prob_triplet__('x1', 't', 'x2', 'f', ['x3'], ['t']))
    assert_almost_equal(0.0, data.__get_cond_prob_triplet__('x1', 't', 'x2', 'f', ['x3'], ['f']))

    assert_almost_equal(0.16666666666666669, data.__get_cond_prob_triplet__('x1', 'f', 'x2', 't', ['x3'], ['t']))
    assert_almost_equal(0.0, data.__get_cond_prob_triplet__('x1', 'f', 'x2', 't', ['x3'], ['f']))
    assert_almost_equal(0.0, data.__get_cond_prob_triplet__('x1', 'f', 'x2', 'f', ['x3'], ['t']))
    assert_almost_equal(1.0, data.__get_cond_prob_triplet__('x1', 'f', 'x2', 'f', ['x3'], ['f']))


@with_setup(setup, teardown)
def test_get_mi():
    """
    Tests getting mutual information.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.19274475702175753, data.__get_mi__('x1', 'x2'))
    assert_almost_equal(0.4228104552401625, data.__get_mi__('x1', 'x3'))
    assert_almost_equal(0.6931471805599453, data.__get_mi__('x1', 'x4'))
    assert_almost_equal(0.4228104552401625, data.__get_mi__('x2', 'x3'))
    assert_almost_equal(0.19274475702175753, data.__get_mi__('x2', 'x4'))
    assert_almost_equal(0.4228104552401625, data.__get_mi__('x3', 'x4'))


@with_setup(setup, teardown)
def test_get_cond_mi():
    """
    Tests getting conditional mutual information.
    :return: None.
    """
    data = DiscreteData(get_good_df())

    assert_almost_equal(0.0, data.__get_cond_mi__('x1', 'x2', ['x3', 'x4']))
    assert_almost_equal(0.0, data.__get_cond_mi__('x1', 'x3', ['x2', 'x4']))
    assert_almost_equal(0.25020121176909393, data.__get_cond_mi__('x1', 'x4', ['x2', 'x3']))

    assert_almost_equal(0.0, data.__get_cond_mi__('x2', 'x1', ['x3', 'x4']))
    assert_almost_equal(0.25020121176909393, data.__get_cond_mi__('x2', 'x3', ['x1', 'x4']))
    assert_almost_equal(0.0, data.__get_cond_mi__('x2', 'x4', ['x1', 'x3']))

    assert_almost_equal(0.0, data.__get_cond_mi__('x3', 'x1', ['x2', 'x4']))
    assert_almost_equal(0.25020121176909393, data.__get_cond_mi__('x3', 'x2', ['x1', 'x4']))
    assert_almost_equal(0.0, data.__get_cond_mi__('x3', 'x4', ['x1', 'x2']))

    assert_almost_equal(0.25020121176909393, data.__get_cond_mi__('x4', 'x1', ['x2', 'x3']))
    assert_almost_equal(0.0, data.__get_cond_mi__('x4', 'x2', ['x1', 'x3']))
    assert_almost_equal(0.0, data.__get_cond_mi__('x4', 'x3', ['x1', 'x2']))


@with_setup(setup, teardown)
def test_get_pairwise_mutual_information():
    """
    Tests get pairwise mutual information.
    :return: List of pairwise mutual information.
    """
    data = DiscreteData(get_good_df())
    observed = sorted(data.get_pairwise_mutual_information(), key=lambda tup: (-tup[2], tup[0], tup[1]))
    expected = [('x1', 'x4'), ('x1', 'x3'), ('x2', 'x3'), ('x3', 'x4'), ('x1', 'x2'), ('x2', 'x4')]

    print(observed)
    print('--')
    print(expected)

    assert len(observed) == len(expected)
    for o, e in zip(observed, expected):
        assert o[0] == e[0]
        assert o[1] == e[1]



from pybbn.graph.potential import PotentialUtil


class EvidenceDistributor(object):
    """
    Evidence distributor. Passes messages using breadth-first-search (BFS).
    Messages are passed from the start clique to the far remote cliques.
    """

    def __init__(self, join_tree, start_clique):
        """
        Ctor.
        :param join_tree: Join tree.
        :param start_clique: Start clique.
        """
        self.join_tree = join_tree
        self.start_clique = start_clique

    def start(self):
        """
        Starts the evidence distribution.
        """
        self.start_clique.mark()
        for sep_set_id in self.join_tree.get_neighbors(self.start_clique.id):
            sep_set = self.join_tree.get_node(sep_set_id)
            for clique_id in self.join_tree.get_neighbors(sep_set_id):
                y = self.join_tree.get_node(clique_id)
                if not y.is_marked():
                    PotentialUtil.pass_single_message(self.join_tree, self.start_clique, sep_set, y)
                    self.__walk__(self.start_clique, sep_set, y)

    def __walk__(self, x, s, y):
        """
        Walks away from the specified node y.
        :param x: Clique.
        :param s: Separation-set.
        :param y: Clique.
        """
        y.mark()
        for sep_set_id in self.join_tree.get_neighbors(y.id):
            sep_set = self.join_tree.get_node(sep_set_id)
            for clique_id in self.join_tree.get_neighbors(sep_set_id):
                clique = self.join_tree.get_node(clique_id)
                if not clique.is_marked():
                    PotentialUtil.pass_single_message(self.join_tree, y, sep_set, clique)
                    self.__walk__(y, sep_set, clique)



from nose import with_setup

from pybbn.graph.potential import Potential, PotentialEntry


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_potential_entry_matches():
    """
    Tests matching potential entry.
    :return: None.
    """
    entry1 = PotentialEntry()
    entry1.add(0, 'on')
    entry1.add(1, 'on')

    entry2 = PotentialEntry()
    entry2.add(0, 'on')
    entry2.add(1, 'on')

    entry3 = PotentialEntry()
    entry3.add(0, 'on')
    entry3.add(1, 'off')

    assert entry1.matches(entry2) == 1
    assert entry1.matches(entry3) == 0


@with_setup(setup, teardown)
def test_potential_entry_duplicate():
    """
    Tests checking for potential duplicates.
    :return: None.
    """
    entry1 = PotentialEntry()
    entry1.add(0, 'on')
    entry1.add(1, 'on')

    entry2 = entry1.duplicate()

    assert entry1.matches(entry2) == 1


@with_setup(setup, teardown)
def test_potential_get_matching_entry():
    """
    Tests get matching entry from potential.
    :return: None.
    """
    entry1 = PotentialEntry().add(0, 'on').add(1, 'on')
    entry2 = PotentialEntry().add(0, 'on').add(1, 'off')
    entry3 = PotentialEntry().add(0, 'off').add(1, 'on')
    entry4 = PotentialEntry().add(0, 'off').add(1, 'off')
    entry5 = PotentialEntry().add(3, 'off').add(1, 'off')

    potential = Potential() \
        .add_entry(entry1) \
        .add_entry(entry2) \
        .add_entry(entry3) \
        .add_entry(entry4)

    assert len(potential.get_matching_entries(entry1)) == 1
    assert len(potential.get_matching_entries(entry2)) == 1
    assert len(potential.get_matching_entries(entry3)) == 1
    assert len(potential.get_matching_entries(entry4)) == 1
    assert len(potential.get_matching_entries(entry5)) == 0


@with_setup(setup, teardown)
def test_str():
    """
    Tests str function.
    :return: None.
    """
    entry1 = PotentialEntry().add(0, 'on').add(1, 'on')
    entry2 = PotentialEntry().add(0, 'on').add(1, 'off')
    entry3 = PotentialEntry().add(0, 'off').add(1, 'on')
    entry4 = PotentialEntry().add(0, 'off').add(1, 'off')

    potential = Potential() \
        .add_entry(entry1) \
        .add_entry(entry2) \
        .add_entry(entry3) \
        .add_entry(entry4)

    assert potential.__str__() == '0=on,1=on|1.0\n0=on,1=off|1.0\n0=off,1=on|1.0\n0=off,1=off|1.0'



from nose import with_setup

from pybbn.generator.bbngenerator import convert_for_drawing
from pybbn.generator.bbngenerator import generate_singly_bbn, generate_multi_bbn, convert_for_exact_inference


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_singly_connected():
    """
    Tests generating singly-connected BBN.
    :return: None.
    """
    g, p = generate_singly_bbn(3)

    assert len(g.nodes) == 3
    assert len(g.edges) == 2


@with_setup(setup, teardown)
def test_multi_connected():
    """
    Tests generating multi-connected BBN.
    :return: None.
    """
    g, p = generate_multi_bbn(4)

    assert len(g.nodes) == 4
    assert len(g.edges) > 1


@with_setup(setup, teardown)
def test_convert_for_exact_inference():
    """
    Tests converting graph and params to a BBN.
    :return: None
    """

    g, p = generate_singly_bbn(2)
    bbn = convert_for_exact_inference(g, p)

    assert len(bbn.nodes) == 2
    assert len(bbn.edges) == 1


@with_setup(setup, teardown)
def test_convert_for_drawing():
    """
    Tests converting BBN to networkx DAG.
    :return: None
    """

    g, p = generate_singly_bbn(2)
    bbn = convert_for_exact_inference(g, p)

    graph = convert_for_drawing(bbn)
    nodes = list(graph.nodes)
    edges = list(graph.edges)

    assert len(nodes) == 2
    assert len(edges) == 1



from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.graph import Ug
from pybbn.graph.node import Clique


class Triangulator(object):
    """
    Triangulator. Triangulates an undirected moralized graph and produces cliques in the process.
    """

    @staticmethod
    def triangulate(m):
        """
        Triangulates the specified moralized graph.
        :param m: Moralized undirected graph.
        :return: Array of cliques.
        """
        cliques = []
        mm = Triangulator.duplicate(m)
        while len(mm.get_nodes()) > 0:
            node_clique = Triangulator.select_node(mm)
            clique = Clique(node_clique.get_bbn_nodes())

            if not Triangulator.is_subset(cliques, clique):
                cliques.append(clique)

            mm.remove_node(node_clique.node.id)

            for edge in node_clique.edges:
                m.add_edge(edge)
                mm.add_edge(edge)

        return cliques

    @staticmethod
    def duplicate(g):
        """
        Duplicates a undirected graph.
        :param g: Undirected graph.
        :return: Undirected graph.
        """
        ug = Ug()
        for node in g.get_nodes():
            ug.add_node(node)
        for edge in g.get_edges():
            ug.add_edge(edge)
        return ug

    @staticmethod
    def select_node(m):
        """
        Selects a clique from the specified graph. Cliques are sorted by number of edges, weight, and id (asc).
        :param m: Graph.
        :return: Clique.
        """
        cliques = []
        for node in m.get_nodes():
            weight = Triangulator.get_weight(node, m)
            edges = Triangulator.get_edges_to_add(node, m)
            neighbors = [m.get_node(neighbor_id) for neighbor_id in m.get_neighbors(node.id)]
            cliques.append(NodeClique(node, neighbors, weight, edges))
        cliques = sorted(cliques, key=lambda x: (len(x.edges), x.weight, x.node.id))
        return cliques[0]

    @staticmethod
    def get_weight(n, m):
        """
        Gets the weight of a BBN node. The weight of a node is the product of the its weight with all its
        neighbors' weight.
        :param n: BBN node.
        :param m: Graph.
        :return: Weight.
        """
        weight = n.get_weight()
        for neighbor_id in m.get_neighbors(n.id):
            neighbor = m.get_node(neighbor_id)
            weight *= neighbor.get_weight()
        return weight

    @staticmethod
    def get_edges_to_add(n, m):
        """
        Gets edges to add.
        :param n: BBN node.
        :param m: Graph.
        :return: Array of edges.
        """
        edges = []
        neighbors = [m.get_node(i) for i in m.get_neighbors(n.id)]
        size = len(neighbors)
        for i in range(size):
            ne1 = neighbors[i]
            for j in range(i + 1, size):
                ne2 = neighbors[j]
                if not m.edge_exists(ne1.id, ne2.id):
                    edges.append(Edge(ne1, ne2, EdgeType.UNDIRECTED))
        return edges

    @staticmethod
    def is_subset(cliques, clique):
        """
        Checks if the specified clique is a subset of the specified list of cliques.
        :param cliques: List of cliques.
        :param clique: Clique.
        :return: A boolean indicating if the clique is a subset.
        """
        for i in range(len(cliques)):
            if cliques[i].is_superset(clique):
                return True
        return False


class NodeClique:
    """
    Node clique.
    """

    def __init__(self, node, neighbors, weight, edges):
        """
        Ctor.
        :param node: BBN node.
        :param neighbors: BBN nodes (neighbors).
        :param weight: Weight.
        :param edges: Edges.
        """
        self.node = node
        self.neighbors = neighbors
        self.weight = weight
        self.edges = edges

    def get_bbn_nodes(self):
        """
        Gets all the BBN nodes in this node clique.
        :return: Array of BBN nodes.
        """
        nodes = [node for node in self.neighbors]
        nodes.append(self.node)
        return nodes



from enum import Enum


class EdgeType(Enum):
    """
    Edge type.
    """
    UNDIRECTED = 1
    DIRECTED = 2


class Edge(object):
    """
    Edge.
    """

    def __init__(self, i, j, type):
        """
        Ctor.
        :param i: Node.
        :param j: Node.
        :param type: Edge type.
        """
        self.i = i
        self.j = j
        self.type = type

    @property
    def key(self):
        """
        Key used for map.
        :return: Key.
        """
        a = min(self.i.id, self.j.id)
        b = max(self.i.id, self.j.id)

        edge = '--'
        if EdgeType.DIRECTED == self.type:
            edge = '->'
            return '{}{}{}'.format(self.i.id, edge, self.j.id)

        return "{}{}{}".format(a, edge, b)

    def __str__(self):
        return self.key


class SepSetEdge(Edge):
    """
    Separation set.
    """

    def __init__(self, i, j):
        """
        Ctor.
        :param i: Node.
        :param j: Node.
        """
        Edge.__init__(self, i, j, EdgeType.UNDIRECTED)

    def __str__(self):
        a = min(self.i.id, self.j.id)
        b = max(self.i.id, self.j.id)

        lhs = self.i.__str__() if a == self.i.id else self.j.__str__()
        rhs = self.j.__str__() if b == self.j.id else self.i.__str__()

        edge = '--'

        return "{}{}{}".format(lhs, edge, rhs)


class JtEdge(Edge):
    """
    Junction tree edge. This is basically a hyper-edge.
    """

    def __init__(self, sep_set):
        """
        Ctor.
        :param sep_set: Separation set.
        """
        Edge.__init__(self, sep_set.left, sep_set.right, EdgeType.UNDIRECTED)
        self.sep_set = sep_set

    def get_lhs_edge(self):
        """
        Gets a JtEdge. e.g. left -- sep_set.
        :return: JtEdge.
        """
        return SepSetEdge(self.sep_set.left, self.sep_set)

    def get_rhs_edge(self):
        """
        Gets a JtEdge. e.g. right -- sep_set.
        :return: JtEdge.
        """
        return SepSetEdge(self.sep_set.right, self.sep_set)

    def __str__(self):
        return '{}--{}--{}'.format(self.sep_set.left.__str__(), self.sep_set.__str__(), self.sep_set.right.__str__())



from pybbn.graph.edge import EdgeType


class Graph(object):
    """
    Graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.nodes = dict()
        self.edges = dict()
        self.map = dict()

    def get_neighbors(self, id):
        """
        Gets the neighbors of the specified node.
        :param id: Node id.
        :return: Set of neighbors of the specified node.
        """
        set1 = set([x for x in self.map[id]])
        set2 = set([x for x in self.map if id in self.map[x]])
        return set1 | set2

    def get_node(self, id):
        """
        Gets the node associated with the specified id.
        :param id: Node id.
        :return: Node.
        """
        return self.nodes[id]

    def get_nodes(self):
        """
        Gets all the nodes.
        :return: List of nodes.
        """
        return [node for node in self.nodes.values()]

    def get_edges(self):
        """
        Gets all the edges.
        :return: List of edges.
        """
        return [edge for edge in self.edges.values()]

    def add_node(self, node):
        """
        Adds a node.
        :param node: Node.
        :return: This graph.
        """
        self.nodes[node.id] = node
        return self

    def add_edge(self, edge):
        """
        Adds an edge.
        :param edge: Edge.
        :return: This graph.
        """
        self.add_node(edge.i)
        self.add_node(edge.j)

        if edge.i.id not in self.map:
            self.map[edge.i.id] = set()
        if edge.j.id not in self.map:
            self.map[edge.j.id] = set()

        if self.__shouldadd__(edge):
            self.edges[edge.key] = edge
            self.map[edge.i.id].add(edge.j.id)
            if EdgeType.UNDIRECTED == edge.type:
                self.map[edge.j.id].add(edge.i.id)
            self.__edge_added__(edge)

        return self

    def __edge_added__(self, edge):
        """
        Callback listener for sub-classes when an edge has been added.
        :param edge: Edge.
        :return: None.
        """
        pass

    def __shouldadd__(self, edge):
        """
        Checks if the specified edge should be added.
        :param edge: Edge.
        :return: A boolean indicating if the edge should be added.
        """
        lhs = edge.i
        rhs = edge.j

        if lhs.id == rhs.id:
            return False

        if EdgeType.UNDIRECTED == edge.type:
            if lhs.id not in self.map[rhs.id] or rhs.id not in self.map[lhs.id]:
                return True
        else:
            if rhs.id not in self.map[lhs.id]:
                return True

        return False

    def edge_exists(self, id1, id2):
        """
        Checks if the specified edge id1 -- id2 exists.
        :param id1: Node id.
        :param id2: Node id.
        :return: A boolean indicating if the specified edge exists.
        """
        if id1 in self.map and id2 in self.map[id1]:
            return True
        if id2 in self.map and id1 in self.map[id2]:
            return True
        return False

    def remove_node(self, id):
        """
        Removes a node from the graph.
        :param id: Node id.
        """
        self.nodes.pop(id, None)
        self.map.pop(id, None)
        for k, v in self.map.items():
            if id in v:
                v.remove(id)

    def __str__(self):
        nodes = str.join('\n', [x.__str__() for x in self.nodes.values()])
        edges = str.join('\n', [x.__str__() for x in self.edges.values()])
        return nodes + '\n' + edges


class Ug(Graph):
    """
    Undirected graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        Graph.__init__(self)



import math

import numpy as np
from nose import with_setup
from nose.tools import assert_almost_equal
from scipy.stats import norm

from pybbn.lg.gaussian import __get_sample__, rnorm, dnorm, rmvnorm, dmvnorm, rcmvnorm, dcmvnorm


def setup():
    """
    Setup.
    :return: None.
    """
    np.random.seed(37)


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_gaussian_sampling():
    """
    Tests Gaussian sampling.
    :return: None.
    """
    mean, stdev = 3, 10
    samples = np.asarray([__get_sample__(mean, stdev) for _ in range(9000)])
    m, s = samples.mean(), math.sqrt(samples.var())
    print('m={}, s={}'.format(m, s))
    assert_almost_equal(mean, m, delta=0.1)
    assert_almost_equal(stdev, s, delta=0.1)


@with_setup(setup, teardown)
def test_rnorm():
    """
    Tests sampling from Gaussian distribution.
    :return: None.
    """
    mean, stdev = 3, 10
    samples = np.asarray(list(rnorm(9000, mean, stdev)))
    m, s = samples.mean(), math.sqrt(samples.var())
    print('m={}, s={}'.format(m, s))
    assert_almost_equal(mean, m, delta=0.1)
    assert_almost_equal(stdev, s, delta=0.1)


@with_setup(setup, teardown)
def test_dnorm():
    """
    Tests computing probabilities of values for a Gaussian distribution.
    :return: None.
    """
    mean, stdev = 3, 10
    samples = np.asarray(list(rnorm(9000, mean, stdev)))

    prob_a = sum(np.log(np.asarray(list(dnorm(samples, mean, stdev)))))
    prob_b = sum(np.log(norm.pdf(samples, mean, stdev)))
    assert_almost_equal(prob_a, prob_b, delta=0.01)


@with_setup(setup, teardown)
def test_rmvnorm():
    """
    Tests sampling from multivariate Gaussian distribution.
    :return: None.
    """
    samples = np.array(list(rmvnorm(
        90000,
        np.array([0.0, 25.0]),
        np.array([
            [1.091385, 1.949606],
            [1.949606, 4.520746]]))
    ))
    means = samples.mean(axis=0)
    x_mean, y_means = means[0], means[1]
    assert_almost_equal(x_mean, 0.0, delta=0.01)
    assert_almost_equal(y_means, 25.0, delta=0.01)


@with_setup(setup, teardown)
def test_dmvnorm():
    """
    Tests computing probabilities of values for a multivariate Gaussian distribution.
    :return: None.
    """
    p1 = sum(np.log(np.array(list(dmvnorm(
        data=np.array([[0, 25], [1, 26], [0, 26]]),
        m=np.array([0.0, 25.0]),
        cov=np.array([
            [1.091385, 1.949606],
            [1.949606, 4.520746]])
    )))))
    assert_almost_equal(p1, -6.938470946099255, places=1)

    p2 = sum(np.log(np.array(list(dmvnorm(
        data=np.array([[0, 25], [1, 26], [0, 27]]),
        m=np.array([0.0, 25.0]),
        cov=np.array([
            [1.091385, 1.949606],
            [1.949606, 4.520746]])
    )))))
    assert_almost_equal(p2, -8.383489807272817, places=1)

    assert p1 > p2


@with_setup(setup, teardown)
def test_rcmvnorm():
    """
    Tests sampling from a conditional multivariate Gaussian distribution.
    :return: None.
    """
    m = np.array(list(rcmvnorm(
        9000,
        np.array([0, 25]),
        np.array([[1.09, 1.95], [1.95, 4.52]]),
        1,
        [0],
        np.array([1])))).mean()
    assert_almost_equal(m, 26.793937408775005, places=1)

    m = np.array(list(rcmvnorm(
        90000,
        np.array([0, 25]),
        np.array([[1.09, 1.95], [1.95, 4.52]]),
        1,
        [0],
        np.array([0])))).mean()
    assert_almost_equal(m, 25.000056049420696, places=1)

    m = np.array(list(rcmvnorm(
        90000,
        np.array([0, 25]),
        np.array([[1.09, 1.95], [1.95, 4.52]]),
        0,
        [1],
        np.array([20])))).mean()
    assert_almost_equal(m, -2.156362075970362, delta=0.01)


@with_setup(setup, teardown)
def test_dcmvnorm():
    """
    Tests computing probabilities of values for a conditional multivariate Gaussian distribution.
    :return: None.
    """
    p1 = sum(np.log(np.array(list(dcmvnorm(
        data=np.array([
            [0, 25],
            [0, 25],
            [0, 24],
            [1, 26],
            [1, 25]
        ]),
        m=np.array([0, 25]),
        cov=np.array([[1.09, 1.95], [1.95, 4.52]]),
        dep=1,
        given=[0]
    )))))
    assert_almost_equal(p1, -7.0162097830845696, places=1)

    p2 = sum(np.log(np.array(list(dcmvnorm(
        data=np.array([
            [0, 25],
            [0, 25],
            [0, 24],
            [1, 25],
            [1, 25]
        ]),
        m=np.array([0, 25]),
        cov=np.array([[1.09, 1.95], [1.95, 4.52]]),
        dep=1,
        given=[0]
    )))))
    assert_almost_equal(p2, -8.227751578381055, places=2)

    assert p1 > p2



from pybbn.graph.potential import PotentialUtil


class PotentialInitializer(object):
    """
    Potential initializer.
    """

    @staticmethod
    def init(bbn):
        """
        Initializes the BBN potentials.
        :param bbn: BBN graph.
        """
        for node in bbn.get_nodes():
            parents = [bbn.get_node(parent_id) for parent_id in bbn.get_parents_ordered(node.id)]
            potential = PotentialUtil.get_potential(node, parents)
            node.potential = potential



import itertools
import json

from pybbn.graph.dag import Bbn
from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.node import BbnNode
from pybbn.graph.variable import Variable


class Factory(object):
    """
    Factory to convert other API BBNs into py-bbn.
    """

    @staticmethod
    def from_libpgm_discrete_json(j):
        """
        Converts a libpgm discrete network as specified by a JSON string into a py-bbn one.
        Look <a href="https://pythonhosted.org/libpgm/unittestdict.html">here</a>.
        :param j: String representing JSON.
        :return: py-bbn BBN.
        """
        return Factory.from_libpgm_discrete_dictionary(json.loads(j))

    @staticmethod
    def from_libpgm_discrete_dictionary(d):
        """
        Converts a libpgm discrete network as specified by a dictionary into a py-bbn one.
        Look <a href="https://pythonhosted.org/libpgm/unittestdict.html">here</a>.
        :param d: A dictionary representing a libpgm discrete network.
        :return: py-bbn BBN.
        """

        class LibpgmBBN(object):
            def __init__(self, V, E, Vdata):
                self.V = V
                self.E = E
                self.Vdata = Vdata

        bn = LibpgmBBN(d['V'], d['E'], d['Vdata'])
        return Factory.from_libpgm_discrete_object(bn)

    @staticmethod
    def from_libpgm_discrete_object(bn):
        """
        Converts a libpgm discrete network object into a py-bbn one.
        :param bn: libpgm discrete BBN.
        :return: py-bbn BBN.
        """
        def get_nodes(bn, domain_spaces=True):
            def get_parent_domains(name, bn):
                parents = bn.Vdata[name]['parents']
                domains = []

                if parents is None or len(parents) == 0:
                    return domains

                for parent in parents:
                    domain = bn.Vdata[parent]['vals'][:]
                    domains.append(domain)
                return domains

            def cross_product(domains):
                products = []

                if domains is None or len(domains) == 0:
                    return products

                for e in itertools.product(*domains):
                    products.append(e)
                return products

            def stringify_cross_product(pa_domains, domain_spaces=True):
                joiner_delim = ', ' if domain_spaces is True else ','
                s = []
                for pa_domain in pa_domains:
                    r = joiner_delim.join(["'{}'".format(v) for v in pa_domain])
                    r = '[{}]'.format(r)
                    s.append(r)
                return s

            def get_cond_probs(name, bn, domain_spaces=True):
                probs = []
                pa_domains = stringify_cross_product(cross_product(get_parent_domains(name, bn)), domain_spaces)
                if len(pa_domains) == 0:
                    probs = bn.Vdata[name]['cprob'][:]
                else:
                    for pa_domain in pa_domains:
                        cprob = bn.Vdata[name]['cprob'][pa_domain]
                        probs.extend(cprob)

                return probs

            nodes = {}
            idx = 0
            for name in bn.V:
                domain = bn.Vdata[name]['vals'][:]
                probs = get_cond_probs(name, bn, domain_spaces)
                node = BbnNode(Variable(idx, name, domain), probs)
                nodes[name] = node
                idx += 1
            return nodes

        def get_edges(bn, nodes):
            edges = []
            for k, v in bn.Vdata.items():
                ch = nodes[k]
                if v['parents'] is not None and len(v['parents']) > 0:
                    parents = [nodes[pa] for pa in v['parents']]
                    for pa in parents:
                        edge = Edge(pa, ch, EdgeType.DIRECTED)
                        edges.append(edge)
            return edges

        nodes = get_nodes(bn)
        edges = get_edges(bn, nodes)

        bbn = Bbn()

        for k, v in nodes.items():
            bbn.add_node(v)

        for e in edges:
            bbn.add_edge(e)

        return bbn



from nose import with_setup

from pybbn.graph.dag import BbnUtil
from pybbn.pptc.initializer import Initializer
from pybbn.pptc.moralizer import Moralizer
from pybbn.pptc.potentialinitializer import PotentialInitializer
from pybbn.pptc.transformer import Transformer
from pybbn.pptc.triangulator import Triangulator


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_initializer():
    """
    Tests initialization.
    :return: None.
    """
    bbn = BbnUtil.get_huang_graph()
    PotentialInitializer.init(bbn)

    ug = Moralizer.moralize(bbn)
    cliques = Triangulator.triangulate(ug)

    join_tree = Transformer.transform(cliques)

    Initializer.initialize(join_tree)

    # assert later
    # for clique in join_tree.get_cliques():
    #     if isinstance(clique, SepSet):
    #         continue
    #     potential = join_tree.potentials[clique.id]
    #     print(clique)
    #     print(potential)
    # assert 1 == 2



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_desc = fh.read()

setup(
    name='pybbn',
    version='0.0.9',
    author='Jee Vang',
    author_email='vangjee@gmail.com',
    packages=find_packages(),
    description='Learning and Inference in Bayesian Belief Networks',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url='https://github.com/vangj/py-bbn',
    keywords=['bayesian', 'belief', 'network', 'exact', 'approximate', 'inference', 'junction', 'tree', 'algorithm',
              'pptc', 'dag', 'gibbs', 'sampling', 'multivariate', 'conditional', 'gaussian', 'linear', 'causal',
              'causality', 'structure', 'parameter'],
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta'
    ]
)



from pybbn.graph.util import IdUtil


class Node(object):
    """
    A node.
    """

    def __init__(self, id):
        """
        :param id: Numeric identifier.
        """
        self.id = id
        self.metadata = dict()

    def add_metadata(self, k, v):
        """
        Adds metadata.
        :param k: Key. Typically a string value.
        :param v: Value. Any object.
        """
        self.metadata[k] = v

    def __str__(self):
        return '{}'.format(self.id)


class BbnNode(Node):
    """
    A BBN node.
    """

    def __init__(self, variable, probs):
        """
        :param variable: A variable.
        :param probs: Array of conditional probabilities.
        """
        Node.__init__(self, variable.id)
        self.variable = variable
        self.probs = probs
        self.potential = None

    def get_weight(self):
        """
        Gets the weight, which is the number of values.
        :return: Weight.
        """
        return len(self.variable.values)

    def __str__(self):
        return '{}|{}|{}'.format(self.id, self.variable.name, str.join(',', self.variable.values))


class Clique(Node):
    """
    A clique.
    """

    def __init__(self, nodes):
        """
        :param nodes: An array of BbnNodes.
        """
        nids = [n.id for n in nodes]
        nids.sort()
        sid = str.join('-', [str(x) for x in nids])
        id = IdUtil.hash_string(sid)
        Node.__init__(self, id)
        self.nodes = nodes
        self.marked = False

    def is_marked(self):
        """
        :return: A boolean indicating if the clique is marked.
        """
        return self.marked

    def mark(self):
        """
        Marks this clique.
        """
        self.marked = True

    def unmark(self):
        """
        Unmarks this clique.
        """
        self.marked = False

    def get_node_ids(self):
        """
        :return: An array of numeric ids of the nodes in this clique.
        """
        return [node.id for node in self.nodes]

    def is_superset(self, that):
        """
        :param that: Clique.
        :return: A boolean indicating if this clique is a superset of the clique passed in.
        """
        s1 = set(self.get_node_ids())
        s2 = set(that.get_node_ids())
        return s1.issuperset(s2)

    def get_weight(self):
        """
        Gets the weight of this clique; the weight is product of the weights of the nodes in this clique.
        :return: Weight.
        """
        weight = 1
        for node in self.nodes:
            weight = weight * node.get_weight()
        return weight

    def contains(self, id):
        """
        :param id: Numeric id.
        :return: A boolean indicating if the specified id exists in this clique.
        """
        result = len([n for n in self.nodes if n.id == id])
        if result > 0:
            return True
        else:
            return False

    def get_sep_set(self, that):
        """
        Creates a separation-set from this node and the one passed in. The separation-set is composed
        of the intersection of the two cliques. If this node has [0, 1, 2] and the node passed in has
        [1, 2, 3], then the separation set will be [1, 2].
        :param that: Clique.
        :return: Separation-set.
        """
        return SepSet(self, that)

    def __str__(self):
        names = sorted([node.variable.name for node in self.nodes])
        return '({})'.format(str.join(',', names))


class SepSet(Clique):
    """
    Separation-set.
    """

    def __init__(self, left, right):
        """
        :param left: Clique.
        :param right: Clique.
        """
        lhs = [x.id for x in left.nodes]
        rhs = [x.id for x in right.nodes]
        intersection = [x for x in lhs if x in rhs]
        nodes = [x for x in left.nodes if x.id in intersection]
        Clique.__init__(self, nodes)
        self.left = left
        self.right = right

    @property
    def cost(self):
        """
        :return: The cost.
        """
        return self.get_cost()

    @property
    def mass(self):
        """
        :return: The mass.
        """
        return self.get_mass()

    def get_cost(self):
        """
        The cost is the sum of the weights of the cliques connected to this separation-set.
        :return: Cost.
        """
        return self.left.get_weight() + self.right.get_weight()

    def get_mass(self):
        """
        The mass is the number of nodes in this separation-set.
        :return: Mass.
        """
        return len(self.nodes)

    def __str__(self):
        return '|{}|'.format(str.join(',', [node.variable.name for node in self.nodes]))



import numpy as np
import pandas as pd
from nose import with_setup
from nose.tools import assert_almost_equal

from pybbn.lg.graph import Dag, Parameters, Bbn


def setup():
    """
    Setup.
    :return: None.
    """
    np.random.seed(37)


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_create_dag():
    """
    Tests creating a basic dag.
    :return: None.
    """
    dag = Dag()
    dag.add_node(0)
    dag.add_node(1)
    dag.add_edge(0, 1)

    assert dag.number_of_nodes() == 2
    assert dag.number_of_edges() == 1
    assert 0 in dag.nodes()
    assert 1 in dag.nodes()
    assert (0, 1) in dag.edges()
    assert 0 == len(dag.parents(0))
    assert 1 == len(dag.children(0))
    assert 1 == len(dag.parents(1))
    assert 0 == len(dag.children(1))
    assert [0, 1] == dag.get_sorted_topology()


@with_setup(setup, teardown)
def test_create_dag_with_cycle():
    """
    Tests creating a DAG with a cycle.
    :return: None.
    """
    try:
        dag = Dag()
        dag.add_node(0)
        dag.add_node(1)
        dag.add_edge(0, 1)
        dag.add_edge(1, 0)

        assert True is False
    except ValueError:
        assert True is True


@with_setup(setup, teardown)
def test_markov_blanket():
    """
    Tests getting Markov blanket.
    :return: None.
    """
    g = Dag()
    g.add_node(0)
    g.add_node(1)
    g.add_node(2)
    g.add_node(3)
    g.add_node(4)
    g.add_edge(0, 1)
    g.add_edge(1, 2)
    g.add_edge(3, 2)

    blanket = g.markov_blanket(1)
    assert 3 == len(blanket)
    assert 0 in blanket
    assert 2 in blanket
    assert 3 in blanket


@with_setup(setup, teardown)
def test_create_parameters():
    """
    Tests creating parameters.
    :return: None.
    """
    try:
        Parameters(np.array([0, 0]), np.array([[0, 0], [0, 0]]))
        assert True is True
    except ValueError:
        assert True is False


@with_setup(setup, teardown)
def test_create_parameter_means_not_eq_cov():
    """
    Tests creating parameters where means is not equal to covariance.
    :return: None.
    """
    try:
        Parameters(np.array([0, 0, 0]), np.array([[0, 0], [0, 0]]))
        assert True is False
    except ValueError:
        assert True is True


@with_setup(setup, teardown)
def test_create_parameter_cov_not_square():
    """
    Tests creating parameters where covariance matrix is not square.
    :return: None.
    """
    try:
        Parameters(np.array([0, 0]), np.array([[0, 0], [0, 0, 0]]))
        assert True is False
    except ValueError:
        assert True is True
    except IndexError:
        assert True is True


@with_setup(setup, teardown)
def test_create_bbn():
    """
    Tests creating a BBN.
    :return: None.
    """
    dag = Dag()
    dag.add_node(0)
    dag.add_node(1)
    dag.add_edge(0, 1)

    params = Parameters(np.array([1, 25]), np.array([[1, 2], [2, 4]]))

    bbn = Bbn(dag, params)
    assert bbn is not None

    try:
        dag = Dag()
        dag.add_node(0)

        params = Parameters(np.array([1, 25]), np.array([[1, 2], [2, 4]]))

        Bbn(dag, params)
        assert True is False
    except ValueError:
        assert True is True

    try:
        dag = Dag()
        dag.add_node(1)
        dag.add_node(2)
        dag.add_edge(1, 2)

        params = Parameters(np.array([1, 25]), np.array([[1, 2], [2, 4]]))

        Bbn(dag, params)
        assert True is False
    except ValueError:
        assert True is True

    try:
        dag = Dag()
        dag.add_node(0)
        dag.add_node(2)
        dag.add_edge(0, 2)

        params = Parameters(np.array([1, 25]), np.array([[1, 2], [2, 4]]))

        Bbn(dag, params)
        assert True is False
    except ValueError:
        assert True is True


@with_setup(setup, teardown)
def test_inference():
    """
    Tests inference.
    :return: None.
    """
    dag = Dag()
    dag.add_node(0)
    dag.add_node(1)
    dag.add_edge(0, 1)

    means = np.array([0, 25])
    cov = np.array([
        [1.09, 1.95],
        [1.95, 4.52]
    ])
    params = Parameters(means, cov)

    bbn = Bbn(dag, params, max_samples=9000, max_iters=1)

    s = bbn.do_inference()
    assert_almost_equal(s[0], 0.0, delta=0.01)
    assert_almost_equal(s[1], 25.0, delta=1.0)
    print(s)

    bbn.set_evidence(0, 1)
    s = bbn.do_inference()
    assert_almost_equal(s[0], 1.0, delta=0.00)
    assert_almost_equal(s[1], 25.0, delta=1.0)
    print(s)
    bbn.clear_evidences()

    bbn.set_evidence(1, 20)
    s = bbn.do_inference()
    assert_almost_equal(s[0], -2.0, delta=0.5)
    assert_almost_equal(s[1], 20.0, delta=0.0)
    print(s)
    bbn.clear_evidences()


@with_setup(setup, teardown)
def test_local_inference():
    """
    Tests local inference.
    :return: None.
    """
    dag = Dag()
    dag.add_node(0)
    dag.add_node(1)
    dag.add_edge(0, 1)

    means = np.array([0, 25])
    cov = np.array([
        [1.09, 1.95],
        [1.95, 4.52]
    ])
    params = Parameters(means, cov)

    bbn = Bbn(dag, params, max_samples=9000, max_iters=1, mb=True)

    s = bbn.do_inference()
    assert_almost_equal(s[0], 0.0, delta=0.01)
    assert_almost_equal(s[1], 25.0, delta=1.0)
    print(s)

    bbn.set_evidence(0, 1)
    s = bbn.do_inference()
    assert_almost_equal(s[0], 1.0, delta=0.00)
    assert_almost_equal(s[1], 25.0, delta=1.0)
    print(s)
    bbn.clear_evidences()

    bbn.set_evidence(1, 20)
    s = bbn.do_inference()
    assert_almost_equal(s[0], -2.0, delta=0.5)
    assert_almost_equal(s[1], 20.0, delta=0.0)
    print(s)
    bbn.clear_evidences()


@with_setup(setup, teardown)
def test_log_proba():
    """
    Tests log probability of data given a BBN.
    :return: None.
    """

    num_samples = 10000

    x0 = 2.0 + np.random.standard_normal(num_samples)
    x1 = 5.0 + 2.0 * x0 + np.random.standard_normal(num_samples)
    x2 = 2.0 + np.random.standard_normal(num_samples)
    x3 = 1.0 + 0.3 * x1 + 0.5 * x2 + np.random.standard_normal(num_samples)
    x4 = 8.0 + 0.9 * x3 + np.random.standard_normal(num_samples)

    df = pd.DataFrame({
        'x0': x0,
        'x1': x1,
        'x2': x2,
        'x3': x3,
        'x4': x4})

    means = np.array(df.mean())
    cov = np.array(df.cov().as_matrix())

    params = Parameters(means, cov)

    dag1 = Dag()
    dag1.add_node(0)
    dag1.add_node(1)
    dag1.add_node(2)
    dag1.add_node(3)
    dag1.add_node(4)
    dag1.add_edge(0, 1)
    dag1.add_edge(1, 3)
    dag1.add_edge(2, 3)
    dag1.add_edge(3, 4)
    bbn1 = Bbn(dag1, params, max_samples=9000, max_iters=1, mb=True)
    lp1 = bbn1.log_prob(df.as_matrix())

    dag2 = Dag()
    dag2.add_node(0)
    dag2.add_node(1)
    dag2.add_node(2)
    dag2.add_node(3)
    dag2.add_node(4)
    dag2.add_edge(0, 1)
    dag2.add_edge(1, 3)
    dag2.add_edge(1, 4)
    dag2.add_edge(2, 3)
    bbn2 = Bbn(dag2, params, max_samples=9000, max_iters=1, mb=True)
    lp2 = bbn2.log_prob(df.as_matrix())

    dag3 = Dag()
    dag3.add_node(0)
    dag3.add_node(1)
    dag3.add_node(2)
    dag3.add_node(3)
    dag3.add_node(4)
    dag3.add_edge(0, 1)
    dag3.add_edge(1, 3)
    dag3.add_edge(3, 4)
    dag3.add_edge(3, 2)
    bbn3 = Bbn(dag3, params, max_samples=9000, max_iters=1, mb=True)
    lp3 = bbn3.log_prob(df.as_matrix())

    dag4 = Dag()
    dag4.add_node(0)
    dag4.add_node(1)
    dag4.add_node(2)
    dag4.add_node(3)
    dag4.add_node(4)
    dag4.add_edge(0, 1)
    dag4.add_edge(2, 3)
    dag4.add_edge(3, 4)
    bbn4 = Bbn(dag4, params, max_samples=9000, max_iters=1, mb=True)
    lp4 = bbn4.log_prob(df.as_matrix())

    print('lp1 {}'.format(lp1))
    print('lp2 {}'.format(lp2))
    print('lp3 {}'.format(lp3))
    print('lp4 {}'.format(lp4))

    assert lp1 > lp2
    assert lp1 > lp3
    assert lp1 > lp4
